/* 
   COPYRIGHT 2011 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - xmp_session.c */
/* rdk@krupczak.org */

/* sessions management code for both managers and agents; IPv6 support
   added */

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <windows.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <errno.h>

#ifndef WIN32
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <sys/utsname.h>
#include <strings.h>
#endif

#include <stdlib.h>

#include <openssl/ssl.h>

#include "xmp.h"
#include "socklib.h"
#include "xmp_vars.h"
#include "xmp_session.h"
#include "korgdebug.h"

/* each function in this module needs to be retrofitted for SSL XXX */

/* ************************************************************* */
/* each separate user of the XMP session library needs
   to have a stats struct declared */

extern xmpStats_t xmpStats;

extern int xmp_session_verbose;

extern SSL_CTX *ctx;

/* ************************************************************* */
/* set the authenUser for this session; can be set multiple times
   and can change during a session if desired; if not set, 'xmpUser' is 
   used 
*/

DLLExport int xmpSessionSetAuthenUser(xmpSession_t *sptr, char *authenUser)
{
     if (sptr == NULL) {
        return -1;
     }
     if (authenUser == NULL) {
        return -1;
     }

     /* dont specifically zero-terminate */
     strncpy(sptr->authenUser,authenUser,sizeof(sptr->authenUser));

     return 1;
}

/* ************************************************************* */
/* 
   send a reply PDU with error given a pretty serious/fatal problem 
   w/o including the original PDU as template; the session contains
   the authenUser to use

*/

DLLExport int xmpSendErrorResponseFatal(xmpSession_t *sptr, int64_t messageID,
                                     xmpErrorStatus_t error, u_int errorIndex)
{
     xmpPdu_t send;
     int needed,ret;

     if ((sptr == NULL) || (error == noError))
        return -1;

     if (xmpIsSessionConnected(sptr) == 0)
        return -1;

     bzero((void *)&send,sizeof(send));
     send.messageID = messageID;
     send.errorStatus = error;
     send.errorIndex = errorIndex;
     send.msgType = Response;

     send.numberVariables = 0;
     send.varList = send.varListTail = NULL;

     needed = 0;
     ret = buildXmpMessage(&send,sptr->sendBuf,sptr->sendBufSize,&needed);

     if ((ret < 0) && (needed != 0)) {
        sptr->sendBuf = realloc(sptr->sendBuf,needed);
        if (sptr->sendBuf == NULL) {
           fprintf(stderr,"xmpSendErrorResponseFatal: send realloc failed\n");
	   sptr->sendBufSize = 0;
           return -1;
	}

        sptr->sendBufSize = needed;
        ret = buildXmpMessage(&send,sptr->sendBuf,sptr->sendBufSize,&needed);
     }
     if (ret < 0) {
        fprintf(stderr,"xmpSendErrorResponseFatal: send failed\n");
        return -1;
     }

     /* send it */
     if ((ret = xmpSendMessage(sptr,sptr->sendBuf,ret)) < 0) {
        return -1;
     }
     xmpStats.outTotalReqVars += send.numberVariables;

     return 1;
}
 
/* ************************************************************* */
/* given some sort of error on input, send back and error message using the
   original as a template for the error response */

DLLExport int xmpSendErrorResponse(xmpSession_t *sptr, xmpPdu_t *in, 
                                   xmpErrorStatus_t error, int errorIndex)
{
     int ret,needed;

     if ((sptr == NULL) || (in == NULL) || (error == noError))
        return -1;

     if (xmpIsSessionConnected(sptr) == 0)
        return -1;

     in->errorStatus = error;
     in->errorIndex = errorIndex;

     needed = 0;
     ret = buildXmpMessage(in,sptr->sendBuf,sptr->sendBufSize,&needed);

     if ((ret < 0) && (needed != 0)) {
        sptr->sendBuf = realloc(sptr->sendBuf,needed);
        if (sptr->sendBuf == NULL) {
           fprintf(stderr,"xmpSendErrorResponse: send realloc failed\n");
	   sptr->sendBufSize = 0;
           return -1;
	}

        sptr->sendBufSize = needed;
        ret = buildXmpMessage(in,sptr->sendBuf,sptr->sendBufSize,&needed);
     }
     if (ret < 0) {
        fprintf(stderr,"xmpSendErrorResponse: buildXmpMessage failed\n");
        xmpStats.outErrors++;
        return -1;
     }

     /* send it */
     if ((ret = xmpSendMessage(sptr,sptr->sendBuf,ret)) < 0) {
        return -1;
     }
     xmpStats.outTotalReqVars += in->numberVariables;

     return 1;
}

/* ************************************************************* */
/* create/initiate/connect an outgoing session with an XMP agent */

xmpSession_t *newXmpSessionConnect6(struct in6_addr ip6addr, u_short port, 
                                    char *authenUser)
{
     xmpSession_t *sp;
     SSL *ssl;
     BIO *sbio;
     int ret;

     if ((sp = newXmpSession6(ip6addr,port,0,0,authenUser)) == NULL) {
        return NULL;
     }
     sp->flags = XMP_SESSION_OUTGOING;

     /* get tcp socket and bind it */
     if ((sp->sockfd = get_tcp6_socket()) < 0) {
        fprintf(stderr,"new_xmp_session_connect: failed to get tcp6 socket\n");
        destroyXmpSession(sp);        
        return NULL;
     }
     if (bind_tcp6_socket(sp->sockfd,in6addr_any,0,0) < 0) {
         destroyXmpSession(sp);
         return NULL;
     }
     /* connect the session now */
     if (connect_tcp6_socket(sp->sockfd,sp->remIp6,sp->remPort) < 0) {
         destroyXmpSession(sp);
         return NULL;
     }

     if (xmp_session_verbose)
        fprintf(stderr,"newXmpSessionConnect: TCP connection established\n");

     /* we're TCP connected and almost ready to go */
     sp->flags |= XMP_SESSION_CONNECTED;

     /* do the SSL connection now */
     ssl = SSL_new(ctx);
     sbio = BIO_new_socket(sp->sockfd,BIO_NOCLOSE);
     SSL_set_bio(ssl,sbio,sbio);
     sp->ssl = (void *)ssl;
     sp->sbio = (void *)sbio;

     /* docs say to check <= 0 */
     if ((ret = SSL_connect(ssl)) <= 0) {
        if (xmp_session_verbose) {
	   fprintf(stderr,"newXmpSessionConnect: SSL conn failed, %d\n",ret);
           fprintf(stderr,"newXmpSessionConnect: SSL error %d\n",
                   SSL_get_error(ssl,ret));
        }
        destroyXmpSession(sp);;
        return NULL;
     }

     if (xmp_session_verbose)
        fprintf(stderr,"newXmpSessionConnect: SSL conn established\n");

     /* check server certificate */
     if (check_certificate(sp) < 0) {
        destroyXmpSession(sp);
        return NULL;
     }

     if (xmp_session_verbose)
        fprintf(stderr,"newXmpSessionConnect: SSL cert verified\n");

     return sp;
}

xmpSession_t *newXmpSessionConnect(u_int ipaddr, u_short port, char *authenUser)
{
     xmpSession_t *sp;
     SSL *ssl;
     BIO *sbio;
     int ret;

     if ((sp = newXmpSession(ipaddr,port,0,0,authenUser)) == NULL) {
        return NULL;
     }
     sp->flags = XMP_SESSION_OUTGOING;

     /* get tcp socket and bind it */
     if ((sp->sockfd = get_tcp_socket()) < 0) {
        fprintf(stderr,"new_xmp_session_connect: failed to get tcp socket\n");
        destroyXmpSession(sp);        
        return NULL;
     }
     if (bind_tcp_socket(sp->sockfd,0,0,0) < 0) {
         destroyXmpSession(sp);
         return NULL;
     }
     /* connect the session now */
     if (connect_tcp_socket(sp->sockfd,sp->remIp,sp->remPort) < 0) {
         destroyXmpSession(sp);
         return NULL;
     }

     if (xmp_session_verbose)
        fprintf(stderr,"newXmpSessionConnect: TCP connection established\n");

     /* we're TCP connected and almost ready to go */
     sp->flags |= XMP_SESSION_CONNECTED;

     /* do the SSL connection now */
     ssl = SSL_new(ctx);
     sbio = BIO_new_socket(sp->sockfd,BIO_NOCLOSE);
     SSL_set_bio(ssl,sbio,sbio);
     sp->ssl = (void *)ssl;
     sp->sbio = (void *)sbio;

     /* docs say to check <= 0 */
     if ((ret = SSL_connect(ssl)) <= 0) {
        if (xmp_session_verbose) {
	   fprintf(stderr,"newXmpSessionConnect: SSL conn failed, %d\n",ret);
           fprintf(stderr,"newXmpSessionConnect: SSL error %d\n",
                   SSL_get_error(ssl,ret));
        }
        destroyXmpSession(sp);;
        return NULL;
     }

     if (xmp_session_verbose)
        fprintf(stderr,"newXmpSessionConnect: SSL conn established\n");

     /* check server certificate */
     if (check_certificate(sp) < 0) {
        destroyXmpSession(sp);
        return NULL;
     }

     if (xmp_session_verbose)
        fprintf(stderr,"newXmpSessionConnect: SSL cert verified\n");

     return sp;
}

/* ************************************************************* */
/* ************************************************************* */

xmpSession_t *newXmpSessionConnect6Timeout(struct in6_addr ip6addr, 
                                           u_short port, 
                                           int timeout, char *authenUser)
{
     xmpSession_t *sp;
     SSL *ssl;
     BIO *bio;

     if ((sp = newXmpSession6(ip6addr,port,0,0,authenUser)) == NULL) {
        return NULL;
     }
     sp->flags = XMP_SESSION_OUTGOING;

     /* get tcp socket and bind it */
     if ((sp->sockfd = get_tcp6_socket()) < 0) {
        fprintf(stderr,"new_xmp_session_connect: failed to get tcp6 socket\n");
        destroyXmpSession(sp);        
        return NULL;
     }
     if (bind_tcp6_socket(sp->sockfd,in6addr_any,0,0) < 0) {
         destroyXmpSession(sp);
         return NULL;
     }
     /* connect the session now */
     if (connect_tcp6_socket_timeout(sp->sockfd,sp->remIp6,sp->remPort,
                                    timeout) < 0) {
         DBG("newXmpSessionConnectTimeout: tcp conn failed\n");
         destroyXmpSession(sp);
         return NULL;
     }

     /* we're connected and ready to go */
     sp->flags |= XMP_SESSION_CONNECTED;

     /* do the SSL connection now */
     ssl = SSL_new(ctx);
     bio = BIO_new_socket(sp->sockfd,BIO_NOCLOSE);
     SSL_set_bio(ssl,bio,bio);
     sp->ssl = (void *)ssl;
     sp->sbio = (void *)bio;

     if (SSL_connect(ssl) <= 0) {
        destroyXmpSession(sp);
        return NULL;
     }

     /* check server certificate */
     if (check_certificate(sp) < 0) {
        destroyXmpSession(sp);
        return NULL;
     }

     if (authenUser != NULL) {
        xmpSessionSetAuthenUser(sp,authenUser);
     }

     return sp;
}

xmpSession_t *newXmpSessionConnectTimeout(u_int ipaddr, u_short port, 
                                          int timeout, char *authenUser)
{
     xmpSession_t *sp;
     SSL *ssl;
     BIO *bio;

     if ((sp = newXmpSession(ipaddr,port,0,0,authenUser)) == NULL) {
        return NULL;
     }
     sp->flags = XMP_SESSION_OUTGOING;

     /* get tcp socket and bind it */
     if ((sp->sockfd = get_tcp_socket()) < 0) {
        fprintf(stderr,"new_xmp_session_connect: failed to get tcp socket\n");
        destroyXmpSession(sp);        
        return NULL;
     }
     if (bind_tcp_socket(sp->sockfd,0,0,0) < 0) {
         destroyXmpSession(sp);
         return NULL;
     }
     /* connect the session now */
     if (connect_tcp_socket_timeout(sp->sockfd,sp->remIp,sp->remPort,
                                    timeout) < 0) {
         DBG("newXmpSessionConnectTimeout: tcp conn failed\n");
         destroyXmpSession(sp);
         return NULL;
     }

     /* we're connected and ready to go */
     sp->flags |= XMP_SESSION_CONNECTED;

     /* do the SSL connection now */
     ssl = SSL_new(ctx);
     bio = BIO_new_socket(sp->sockfd,BIO_NOCLOSE);
     SSL_set_bio(ssl,bio,bio);
     sp->ssl = (void *)ssl;
     sp->sbio = (void *)bio;

     if (SSL_connect(ssl) <= 0) {
        destroyXmpSession(sp);
        return NULL;
     }

     /* check server certificate */
     if (check_certificate(sp) < 0) {
        destroyXmpSession(sp);
        return NULL;
     }

     if (authenUser != NULL) {
        xmpSessionSetAuthenUser(sp,authenUser);
     }

     return sp;
}

/* ************************************************************* */
/* ************************************************************* */
/* 
   initialize an incoming IPv4 or IPv6 session structure possibly already
   TCP connected by caller ; caller does SSL connect or accept
   and TCP connect if necessary
*/

xmpSession_t *newXmpSession6(struct in6_addr ip6addr, u_short port, int sock, 
                             int flags, char *authenUser)
{
     xmpSession_t *sp;

     sp = (xmpSession_t *)malloc(sizeof(xmpSession_t));
     if (sp == NULL) {
        perror("malloc of xmp session failed");
        return NULL;
     }

     bzero(sp,sizeof(xmpSession_t));
     sp->remIp6 = ip6addr;
     sp->remPort = port;
     sp->createTime = time(NULL);
     sp->sockfd = sock;
     sp->flags = flags;
     sp->ipVersion = 6;

     sp->sendBuf = (u_char *)malloc(128000);
     if (sp->sendBuf != NULL) {
        bzero((void *)sp->sendBuf,128000);
        sp->sendBufSize = 128000;
     }
     sp->recvBuf = (u_char *)malloc(128000);
     if (sp->recvBuf != NULL) {
        bzero((void *)sp->recvBuf,128000);
        sp->recvBufSize = 128000;
     }
 
     /* set to caller's thread; caller can change thread if
        desired */
     sp->thr = pthread_self(); 

     /* dont do SSL setup; caller will do so */

     if (authenUser != NULL) {
        xmpSessionSetAuthenUser(sp,authenUser);
     }

     sp->xmpVersion = 1;

     return sp;
}

xmpSession_t *newXmpSession(u_int ipaddr, u_short port, int sock, 
                            int flags, char *authenUser)
{
     xmpSession_t *sp;

     sp = (xmpSession_t *)malloc(sizeof(xmpSession_t));
     if (sp == NULL) {
        perror("malloc of xmp session failed");
        return NULL;
     }

     bzero(sp,sizeof(xmpSession_t));
     sp->remIp = ipaddr;
     sp->remPort = port;
     sp->createTime = time(NULL);
     sp->sockfd = sock;
     sp->flags = flags;
     sp->ipVersion = 4;

     sp->sendBuf = (u_char *)malloc(128000);
     if (sp->sendBuf != NULL) {
        bzero((void *)sp->sendBuf,128000);
        sp->sendBufSize = 128000;
     }
     sp->recvBuf = (u_char *)malloc(128000);
     if (sp->recvBuf != NULL) {
        bzero((void *)sp->recvBuf,128000);
        sp->recvBufSize = 128000;
     }
 
     /* set to caller's thread; caller can change thread if
        desired */
     sp->thr = pthread_self(); 

     /* dont do SSL setup; caller will do so */

     if (authenUser != NULL) {
        xmpSessionSetAuthenUser(sp,authenUser);
     }

     sp->xmpVersion = 1;

     return sp;
}

/* ************************************************************* */
/* close socket, free structure, destry session */

int destroyXmpSession(xmpSession_t *sptr)
{
     if (sptr == NULL)
        return -1;

     /* if we are acting as a daemon, then 0,1,2 fds will be closed
        so there is a chance that fd=0 will be used by a connection
        so close sockfd unless its equal to -1 (UNIX) 
      */

     /* should probably do an SSL shutdown prior to freeing everything */
     if (sptr->ssl)
        SSL_shutdown(sptr->ssl);

#ifndef WIN32
     if (sptr->sockfd >= 0)
        close(sptr->sockfd);
#else
     closesocket(sptr->sockfd);
#endif

     if (sptr->sendBuf != NULL)
        free(sptr->sendBuf);
     if (sptr->recvBuf != NULL)
        free(sptr->recvBuf);

     /* free SSL related structures; SSL_free() frees BIOs, cryptos
        ciphers, etc. */
     if (sptr->ssl != NULL) {
        SSL_free((SSL *)sptr->ssl);
     }

     bzero((void *)sptr,sizeof(xmpSession_t *));

     free(sptr);

     return 1;
}

/* ************************************************************* */
/* ************************************************************* */
/* given an existing session, TCP connect/re-connect with peer
   this function is current unused and needs to be re-evaluated
   for SSL prior to it being used XXX 
*/

int xmpConnect6Session(xmpSession_t *sptr) { }

int xmpConnectSession(xmpSession_t *sptr)
{
     SSL *ssl;
     BIO *sbio;

     if (sptr == NULL)
        return -1;

     if (sptr->sockfd > 0) {
#ifndef WIN32
        close(sptr->sockfd);
#else
        closesocket(sptr->sockfd);
#endif
     }

     /* unset connected flag */
     sptr->flags &= ~(XMP_SESSION_CONNECTED);

     /* get tcp socket and bind it */
     if ((sptr->sockfd = get_tcp_socket()) < 0) {
        fprintf(stderr,"xmpConnectSession: failed to get tcp socket\n");
        return -1;
     }
     if (bind_tcp_socket(sptr->sockfd,0,0,0) < 0) {
         return -1;
     }
     /* connect the session now */
     if (connect_tcp_socket(sptr->sockfd,sptr->remIp,sptr->remPort) < 0) {
         fprintf(stderr,"xmpConnectSession: failed to connect\n");
         return -1;
     }

     /* we're connected and ready to go */
     sptr->flags |= XMP_SESSION_CONNECTED;

     /* SSL connect ? */
     ssl = SSL_new((SSL_CTX *)ctx);
     sbio = BIO_new_socket(sptr->sockfd,BIO_NOCLOSE);
     SSL_set_bio(ssl,sbio,sbio);
     sptr->ssl = (void *)ssl;
     sptr->sbio = (void *)sbio;

     if (SSL_connect(ssl) <= 0) {
        destroyXmpSession(sptr);
        return -1;
     }

     /* check server certificate */
     if (check_certificate(sptr) < 0) {
        destroyXmpSession(sptr);
        return -1;
     }

     return 1;
}

/* ************************************************************* */
/* given an established session, close the tcp connection
   and SSL connection associated with it
*/

int xmpDisconnectSession(xmpSession_t *sptr)
{

     if (sptr == NULL)
        return -1;

     /* SSL disconnect? */
     if (sptr->ssl != NULL) {
        SSL_shutdown((SSL *)sptr->ssl);
     }

     if (sptr->sockfd > 0) {
#ifndef WIN32
        close(sptr->sockfd);
#else
        closesocket(sptr->sockfd);
#endif
     }

     /* unset connected bit */
     if (sptr->flags | XMP_SESSION_CONNECTED) {
        sptr->flags &= ~(XMP_SESSION_CONNECTED);
        sptr->flags &= ~(XMP_SESSION_AUTHEN);
     }

     return 1;
}

/* ************************************************************* */
/* given an XML message, send it to peer via established session */

int xmpSendMessage(xmpSession_t *sptr, u_char *buf, int buflen)
{
     xmpWireHdr_t wh;
     int ret;

     if ((sptr == NULL) || (buf == NULL) || (buflen < 5))
        return -1;

     if ((sptr->flags & XMP_SESSION_CONNECTED) == 0) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpSendMessage: not connected\n");
        return -1;
     }
     if (sptr->sockfd == 0) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpSendMessage: no socket\n");
        return -1;
     }
     if (sptr->ssl == NULL) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpSendMessage: no SSL connection\n");
        return -1;
     }

     /* setup and send the wire header */
     bzero((void *)&wh,sizeof(wh));
     snprintf(wh.msgLenStr,12,"%d",buflen);
     snprintf(wh.xmpVersion,4,"%d",XMP_VERSION);
     if (sptr->authenUser[0] != 0)
        snprintf(wh.authenUser,16,"%s",sptr->authenUser);
     else 
        snprintf(wh.authenUser,16,"%s","xmpUser");

     if ((ret = SSL_write((SSL *)sptr->ssl,(void *)&wh,sizeof(wh))) <= 0) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpSendMessage: send wirehdr failed, %d\n",errno);
        xmpStats.outErrors++;
        return -1;
     }

     /* send the actual message */

     if ((ret = SSL_write((SSL *)sptr->ssl,(void *)buf,buflen)) <= 0) {
        xmpStats.outErrors++;
        if (xmp_session_verbose)
           fprintf(stderr,"xmpSendMessage: send failed, %d\n",errno);
        return -1;
     }

     xmpStats.outBytes += buflen;
     xmpStats.outMessages++;

     sptr->lastActivity = time(NULL);

     return 1;
}

/* ************************************************************* */
/* wait-for and receive a XML message from peer via established session
   write the received message into the session's recvBuf reallocing
   if necessary ; return the number of bytes in xml message
*/

int xmpRecvMessage(xmpSession_t *sptr)
{
     xmpWireHdr_t wh, *whptr;
     u_char *bufptr;
     int ret,nread;
     int msgLen,i;
     char aStr[18];
     char wireBuf[64];
     SSL *ssl;
     int xmpVersion;
     
     if (sptr == NULL)
        return -1;

     if ((sptr->flags & XMP_SESSION_CONNECTED) == 0) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpRecvMessage: not connected\n");
        return -1;
     }
     if (sptr->sockfd == 0) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpRecvMessage: no socket\n");
        return -1;
     }
     ssl = (SSL *)sptr->ssl;
     if (ssl == NULL) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpRecvMessage: no SSL connection\n");
        return -1;
     }

     /* read in the wireheader waiting until we've read the entire structure */
     bzero((void *)&wh,sizeof(wh)); /* make valgrind happy? */
     ret = 0; nread = 0;
     bufptr = (u_char *)&wh;
     bufptr = (u_char *)wireBuf;
     for (nread = 0; nread < sizeof(wh); nread += ret) {
         ret = SSL_read(ssl,bufptr+nread,sizeof(wh)-nread);
         if (ret <= 0)
	    break;
     }
     if (ret == 0) {
        if (xmp_session_verbose)
	   fprintf(stderr,"xmpRecvMessage: SSL shutdown during wirehdr read\n");
        xmpStats.inErrors++;
        return -1;
     }
     if (ret < 0) {
        if (xmp_session_verbose)
	   fprintf(stderr,"xmpRecvMessage: SSL read error, %d\n",
                   SSL_get_error(ssl,ret));
        xmpStats.inErrors++;
        return -1;
     }

     DBG1("xmpRecvMessage: read %d bytes of wire header\n",nread);

     /* decode wire header */
     wireBuf[nread] = 0;
     whptr = (xmpWireHdr_t *)wireBuf;

     /* examine XMP version */
     strncpy(aStr,whptr->xmpVersion,4);
     aStr[4] = 0;
     sptr->xmpVersion = atoi(aStr);

     /* if xmp version is not what we expect, punt now */
     if (sptr->xmpVersion != XMP_VERSION) {
        xmpStats.inErrors++;
        return -1;
     }

     /* figure out length of message from wireheader */ 
     strncpy(aStr,whptr->msgLenStr,12);
     aStr[12] = 0;
     msgLen = atoi(aStr);

     /* examine authenUser and place in session */
     strncpy(sptr->authenUser,whptr->authenUser,16);
     sptr->authenUser[16] = 0;

     DBG3("xmpRecvMessage: authenUser:'%s', len:%d version:%d\n",
	  sptr->authenUser,msgLen,sptr->xmpVersion);

     DBG1("xmpRecvMessage: msgLen is %d\n",msgLen);

     if (msgLen > sptr->recvBufSize) {
        if (xmp_session_verbose)
           fprintf(stderr,"xmpRecvMessage: need more space!\n");
        sptr->recvBuf = realloc(sptr->recvBuf,msgLen+2);
        if (sptr->recvBuf == NULL) {
	   sptr->recvBufSize = 0;
           return -1;
        }
        sptr->recvBufSize = msgLen+2;
     }

     /* now, read in the actual XMP message */
     ret = 0; nread = 0; 
     bufptr = sptr->recvBuf;
 
     for (nread = 0; nread < msgLen; nread += ret) {
         ret = SSL_read(ssl,bufptr+nread,msgLen-nread);
         if (ret <= 0)
	    break;
     }
     if (ret == 0) {
        xmpStats.inErrors++;
        fprintf(stderr,"xmpRecvMessage: shutdown during SSL read of msg\n");
        return -1;
     }
     if (ret < 0) {
        if (xmp_session_verbose)
	   fprintf(stderr,"xmpRecvMessage: SSL read error, %d\n",
                   SSL_get_error(ssl,ret));
        xmpStats.inErrors++;
        return -1;
     }

     /* put a zero termination after data since what we are finding
        is that xml parser may choke on extra content that may be in 
        buffer from previous reads/parses even though we pass in a specific 
        length; we put it at index=nread since the data is in the array
        at locations 0..nread-1 */

     if (nread < sptr->recvBufSize)
        sptr->recvBuf[nread] = 0;

     if (nread != msgLen) {
        fprintf(stderr,"xmpRecvMessage: nread:%d != msgLen:%d\n",nread,msgLen);
     }

     DBG2("xmpRecvMessage: received %d, %d bytes\n",nread,strlen(sptr->recvBuf));
     xmpStats.inBytes+= nread;
     xmpStats.inMessages++;

     return nread;
}

/* ************************************************************* */
/* exported function to send query, receive reply over existing
   connection */

DLLExport int xmpTransaction(xmpSession_t *sptr, xmpPdu_t *s, xmpPdu_t *r)
{
     return xmpSendQueryReceiveReply(sptr,s,r);
}

/* ************************************************************* */
/* send an XML message to peer and wait for XML reply XXX */

int xmpSendQueryReceiveReplyXml(xmpSession_t *sptr, u_char *sbuf, int slen,
                                u_char **rbuf, int *maxrlen)
{
     if ((sptr == NULL) || (sbuf == NULL) || (rbuf == NULL) 
         || (maxrlen == NULL))
        return -1;

     if ((sptr->flags & XMP_SESSION_CONNECTED) == 0) {
        fprintf(stderr,"xmpSendQueryReceiveReply: not connected\n");
        return -1;
     }
     if ((sptr->sockfd == 0)) {
        fprintf(stderr,"xmpSendQueryReceiveReply: no socket\n");
        return -1;
     }

     /* send the query */

     /* receive reply */

     /* if we need to re-alloc space for reply, do so now */

     /* receive the reply message */

     return 1;
}

/* ************************************************************* */
/* given internal-format PDUs, encode and send to peer and then
   wait for reply, decode, return to caller */

DLLExport int xmpSendQueryReceiveReply(xmpSession_t *sptr, xmpPdu_t *s,
                                       xmpPdu_t *r)
{
     int needed,ret;

     if ((sptr == NULL) || (s == NULL) || (r == NULL))
        return -1;

     if (xmpIsSessionConnected(sptr) == 0) {
        return -1;
     }
     if ((sptr->sockfd == 0)) {
        return -1;
     }

     /* take send PDU and convert to XML */
     needed = 0;
     ret = buildXmpMessage(s,sptr->sendBuf,sptr->sendBufSize,&needed);
     if ((ret < 0) && (needed != 0)) {

        sptr->sendBuf = realloc(sptr->sendBuf,needed+2);
        if (sptr->sendBuf == NULL) {
           fprintf(stderr,"xmpSendQueryReceiveReply: send realloc failed\n");
	   sptr->sendBufSize = 0;
           return -1;
	}

        sptr->sendBufSize = needed+2;
        needed = 0;
        ret = buildXmpMessage(s,sptr->sendBuf,sptr->sendBufSize,&needed);
     }
     if (ret < 0) {
        fprintf(stderr,"xmpSendQueryReceiveReply: build failed failed\n");
        return -1;
     }

     /* send it */
     if ((ret = xmpSendMessage(sptr,sptr->sendBuf,ret)) < 0) {
        return -2;
     }

     xmpStats.outTotalReqVars += s->numberVariables;

     switch (s->msgType) {
       case Response: xmpStats.outResponses++; break;
       case GetRequest: xmpStats.outGets++; break;
       case SetRequest: xmpStats.outSets++; break;
       case SelectTableRequest: xmpStats.outSelects++; break;
       case InsertTableRequest: xmpStats.outInserts++; break;
       case DeleteTableRequest: xmpStats.outDeletes++; break;
       case UpdateTableRequest: xmpStats.outUpdates++; break;
       case Trap: xmpStats.outTraps++; break;

       /* dont forget Information!! */
       case Information: xmpStats.outInformation++; break;

       case invalidMsgType:
       default:
          break;
     }

     /* wait for and receive reply */
     if ((ret = xmpRecvMessage(sptr)) < 0) {
        fprintf(stderr,
                "xmpSendQueryReceiveReply: xmpRecvMessage from 0x%x failed\n",
                sptr->remIp);
        return -1;
     }

     /* convert to receive PDU located in pdu 'r' */
     ret = parseXmpMessage(sptr->recvBuf,ret,r);
     if (ret < 0) {

        fprintf(stderr,"failed to parse:\n'%s'\n",sptr->recvBuf);

     }

     switch (r->msgType) {
       case Response: xmpStats.inResponses++; break;
       case GetRequest: xmpStats.inGets++; break;
       case SetRequest: xmpStats.inSets++; break;
       case SelectTableRequest: xmpStats.inSelects++; break;
       case InsertTableRequest: xmpStats.inInserts++; break;
       case DeleteTableRequest: xmpStats.inDeletes++; break;
       case UpdateTableRequest: xmpStats.inUpdates++; break;
       case Trap: xmpStats.inTraps++; break;
       case Information: xmpStats.inInformation++; break;
       case invalidMsgType: break;
     }

     return ret;
}

/* ************************************************************* */

DLLExport int xmpIsSessionConnected(xmpSession_t *sptr)
{
     if (sptr == NULL)
        return 0;

     if ((sptr->flags & XMP_SESSION_CONNECTED) == 0)
        return 0;

     /* check SSL connection also XXX */

     return 1;
}

DLLExport int xmpIsSessionAuthenticated(xmpSession_t *sptr)
{
     if (sptr == NULL)
        return 0;

     if ((sptr->flags & XMP_SESSION_AUTHEN) == 0)
        return 0;

     return 1;
}

/* ************************************************************* */
/* get a comman-separated list of plugins,versions from peer */

DLLExport int xmpGetPeerPluginList(xmpSession_t *sptr,char *pluginlist,int max)
{
     xmpPdu_t query, reply;
     xmpVariable_t var1, *varptr;
     int ret, listlen;
     char *cptr;

     if ((sptr == NULL) || (pluginlist == NULL) || (max < 5)) {
        return -1;
     }

     pluginlist[0] = 0;

     if (xmpIsSessionConnected(sptr) == 0)
        return -1;

     bzero((void *)&query,sizeof(query));
     query.msgType = SelectTableRequest;
     query.messageID = (int64_t)time(NULL);
     query.errorStatus = noError;
     query.errorIndex = 0;

     sprintf(query.tableMib,"core");
     sprintf(query.tableName,"pluginTable");
     sprintf(query.tableKey,"*");

     query.numberVariables = 1;
     query.varList = &var1;
     query.varListTail = NULL;
     bzero((void *)&var1,sizeof(var1));
     var1.next = NULL;
     sprintf(var1.varName.mibName,"core");
     sprintf(var1.varName.objName,"pluginDescr");
     var1.varSyntax = nullSyntax;
     var1.valueStrLen = 0;

     bzero((void *)&reply,sizeof(reply));

     if ((ret = xmpSendQueryReceiveReply(sptr,&query,&reply)) < 0) {
        /* fprintf(stderr,"getPeerPluginList: query failed\n"); */
        ;
     }

     /* pull out each descr */
     if ((reply.errorStatus == noError) && (reply.numberVariables >= 1)) {

         varptr = reply.varList;
         listlen = 0;
         pluginlist[0] = 0;

         while (varptr != NULL) {

           /* take out any ',' in module description */
           if ((cptr = strchr(varptr->valueStr,',')) != NULL)
 	      *cptr = ' ';

           strncat(pluginlist,varptr->valueStr,
                   max-(listlen+varptr->valueStrLen));

	   listlen += varptr->valueStrLen;

           varptr = varptr->next;
           if ((varptr != NULL) && (listlen+2 <= max)) {
              strcat(pluginlist,",");
              listlen += 1;
           }
         }

     } /* if reply variables */

     /* cleanup reply so malloc'd space gets freed */
     xmpPduCleanup(&reply);

     return ret;
}

/* ************************************************************* */
/* get peer/agent version, sysDescr, sysUpTime, sysUpTimeStr
   port, homedir XXX */

DLLExport int xmpGetPeerInfo(xmpSession_t *sptr, char *vers, char *descr,
                             char *sysUptime, char *port, char *homedir,
                             int max)
{
     xmpPdu_t query, reply;
     xmpVariable_t *vars, *varptr;
     int ret;

     if ((sptr == NULL) || (vers == NULL) || (descr == NULL))
        return -1;
     if ((sysUptime == NULL) || (port == NULL) || (homedir == NULL))
        return -1;
     if (max < 5)
        return -1;

     if (xmpIsSessionConnected(sptr) == 0) {
        return -1;
     }

     bzero((void *)&query,sizeof(query));
     query.msgType = GetRequest;
     query.messageID = (int64_t)time(NULL);
     query.errorStatus = noError;
     query.errorIndex = 0;

     /* set up the variables */
     vars = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)*5);
     if (vars == NULL) {
        return -1;
     }

     bzero((void *)vars,5*sizeof(xmpVariable_t));
     query.varList = &vars[0];
     query.varListTail = &vars[4];

     vars[0].next = &vars[1];
     sprintf(vars[0].varName.mibName,"core");
     sprintf(vars[0].varName.objName,"sysDescr");
     vars[0].varSyntax = nullSyntax;
     vars[0].valueStrLen = 0;

     vars[1].next = &vars[2];
     sprintf(vars[1].varName.mibName,"core");
     sprintf(vars[1].varName.objName,"xmpdVersion");
     vars[1].varSyntax = nullSyntax;
     vars[1].valueStrLen = 0;

     vars[2].next = &vars[3];
     sprintf(vars[2].varName.mibName,"core");
     sprintf(vars[2].varName.objName,"sysUpTimeStr");
     vars[2].varSyntax = nullSyntax;
     vars[2].valueStrLen = 0;

     vars[3].next = &vars[4];
     sprintf(vars[3].varName.mibName,"core");
     sprintf(vars[3].varName.objName,"xmpdHomeDir");
     vars[3].varSyntax = nullSyntax;
     vars[3].valueStrLen = 0;

     /* sorta silly; we know it already if we are talking to agent  */
     vars[4].next = NULL;
     sprintf(vars[4].varName.mibName,"core");
     sprintf(vars[4].varName.objName,"xmpdPort"); 
     vars[4].varSyntax = nullSyntax;
     vars[4].valueStrLen = 0;

     /* send the query, wait for response */
     if ((ret = xmpSendQueryReceiveReply(sptr,&query,&reply)) < 0) {
        /* fprintf(stderr,"getPeerInfo: query failed\n");  */
        ;
     }

     /* pull out the answers */

     /* shouldnt port be an int? XXX */

     sysUptime[0] = 0; homedir[0] = 0; vers[0] = 0; descr[0] = 0;

     if ((reply.errorStatus == noError) && (reply.numberVariables == 5)) {

        varptr = reply.varList;
        if (varptr != NULL) {
	   /* XXX */
        }
     }

     /* cleanup reply so malloc'd space gets freed */
     /* we dont do pduCleanup on query but free the vars ourself */

     free(vars);
     xmpPduCleanup(&reply);

     return ret;
}

/* ************************************************************* */
/* given a connection, get the peer's version and sysDescr */

DLLExport int xmpGetPeerVersion(xmpSession_t *sptr,char *vers, char *sysDescr,
                                int max)
{
     xmpPdu_t query, reply;
     xmpVariable_t var1,var2, *varptr;
     int copyLen,ret;

     if ((sptr == NULL) || (vers == NULL) || (sysDescr == NULL))
        return -1;
     if (max < 5)
        return -1;

     if (xmpIsSessionConnected(sptr) == 0) {
        return -1;
     }

     bzero((void *)&query,sizeof(query));
     query.msgType = GetRequest;
     query.messageID = (int64_t)time(NULL);
     query.errorStatus = noError;
     query.errorIndex = 0;

     query.numberVariables = 2;
     query.varList = &var1;
     query.varListTail = &var2;
     bzero((void *)&var1,sizeof(var1));
     bzero((void *)&var2,sizeof(var2));
     var1.next = &var2;
     sprintf(var1.varName.mibName,"core");
     sprintf(var1.varName.objName,"sysDescr");
     var1.varSyntax = nullSyntax;
     var1.valueStrLen = 0;

     var2.next = NULL;
     sprintf(var2.varName.mibName,"core");
     sprintf(var2.varName.objName,"xmpdVersion");
     var2.varSyntax = nullSyntax;
     var2.valueStrLen = 0;
  
     bzero((void *)&reply,sizeof(reply));

     vers[0] = 0;
     sysDescr[0] = 0;

     if ((ret = xmpSendQueryReceiveReply(sptr,&query,&reply)) < 0) {
        /* fprintf(stderr,"getPeerVersion: query failed\n");  */
        ;
     }

     /* pull out version, sysDescr */
     if ((reply.errorStatus == noError) &&
	 (reply.numberVariables == 2)) {

         varptr = reply.varList;
  
         if (varptr != NULL) {
	    copyLen = varptr->valueStrLen + 1; /* plus \0 */
            if (copyLen > max)
 	       copyLen = max;
            strncpy(sysDescr,varptr->valueStr,copyLen);
            sysDescr[copyLen-1] = 0;
            varptr = varptr->next;
         }

         if (varptr != NULL) {
 	    copyLen = varptr->valueStrLen + 1; /* plus \0 */
            if (copyLen > max)
 	       copyLen = max;
            strncpy(vers,varptr->valueStr,copyLen);
            vers[copyLen-1] = 0;
            varptr = varptr->next;
         }
     }

     /* cleanup reply so malloc'd space gets freed */
     xmpPduCleanup(&reply);

     /* if we established session but query failed due to something like
        authorizationFailure, we still return no error but empty 
        version XXX rdk 2/21/2012 */

     return ret;
}

/* ************************************************************* */
