/* 
   COPYRIGHT 2008 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - xmp_users.c */
/* rdk@krupczak.org */

/* XMP protocol users */

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <windows.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>

#include "xmp.h"
#include "xmp_session.h"
#include "xmp_users.h"
#include "linkedlist.h"

static linkedList_t *userList = NULL;

#ifdef WIN32
#define strcasecmp  _stricmp
#endif

/* ************************************************************* */

/* compare by username, not proto ops */
int userCompare(xmpAuthenUser_t *a, xmpAuthenUser_t *b)
{
    if ((a == NULL) || (b == NULL))
        return -1;

    if (strcmp(a->authenUser,b->authenUser) == 0)
       return 0;

     return 1;
}

int userPrint(xmpAuthenUser_t *u)
{
     if (u == NULL)
        return -1;

     fprintf(stderr,"User:'%s' proto:0x%02x mibs:'%s' %d %d\n",
             u->authenUser,u->protocolOperations,
             u->allowedMibs,
             u->valid,
             u->invalid);

     return 1;
}

int userFree(xmpAuthenUser_t *u)
{
     if (u == NULL)
        return -1;

     free(u);

     return 1;
}


/* ************************************************************* */

int initUsers()
{
     userList = newLinkedList(sizeof(xmpAuthenUser_t),
                              userCompare,
                              userPrint,
                              userFree,
                              "userList",1);

     xmpAddUser("xmpUser",GetRequest|SelectTableRequest|Response|Trap,
                "core,mib2");

     return 0;
}

/* ************************************************************* */

int xmpAddUser(char *authenUser, u_int protoOps, char *mibList)
{
     xmpAuthenUser_t u, *user;
     linkedElement_t *e;

     if ((authenUser == NULL) || (protoOps == 0) || (mibList == NULL))
        return -1;
     
     bzero((void *)&u,sizeof(u));

     strncpy(u.authenUser,authenUser,16);
     u.authenUser[16] = 0;

     u.protocolOperations = protoOps;

     strncpy(u.allowedMibs,mibList,255);
     u.allowedMibs[255] = 0;

     if ((e = findElement(userList,(void *)&u)) != NULL) {
        fprintf(stderr,"addUser: duplicate username '%s'\n",authenUser);
        return -1;
     }

     if ((user = (xmpAuthenUser_t *)malloc(sizeof(xmpAuthenUser_t))) == NULL) {
        fprintf(stderr,"addUser: malloc failed, %d\n",errno);
        return -1;
     }

     bcopy((void *)&u,(void *)user,sizeof(u));

     insertHead(userList,(void *)user);

     return 1;

}

/* ************************************************************* */

int dumpUsers()
{
     return traverseList(userList);
}

/* ************************************************************* */
/* validate PDU against authenUser permissions; return True(1)
   if acceptable, False(0) if unacceptable; caller still
   has to check against SSL authentication */

int validatePduPermissions(xmpSession_t *sptr, xmpPdu_t *pdu)
{
     static xmpAuthenUser_t *users = NULL;
     static int numberUsers = 0;
     int i;
     xmpVariable_t *varptr;

     /* set up a cache of users so that we dont have to keep copying
        the list into an array nor do we have to keep locking list */

     if ((numberUsers == 0) || (users == NULL)) {
        users = (xmpAuthenUser_t *)toArrayCopy(userList,&numberUsers);
     }
     if ((users == NULL) || (users == NULL))
        return 0;

     /* match authenUser, protocol op, and mib name */

     for (i=0; i<numberUsers; i++) {

         /* match authenUser */
         if (strcasecmp(sptr->authenUser,users[i].authenUser) == 0) {

	    /* check protocol operation */
	    if (pdu->msgType & users[i].protocolOperations) {

	       /* if all MIBs, we good to go */
	       if (strcasecmp(users[i].allowedMibs,XMP_ALL_MIBS) == 0) {
	 	  return 1;
               }

	       /* check MIB if table */
	       if ((pdu->msgType > SetRequest) && (pdu->msgType < Trap)) {
		  if (strstr(users[i].allowedMibs,pdu->tableMib))
		      return 1;
 	       }
               else {
                  /* check scalars to make sure they are acceptable */
	 	  int all = 1;
                  varptr = pdu->varList;
                  while (varptr != NULL) {
		    if (strstr(users[i].allowedMibs,
                               varptr->varName.mibName) == NULL) { 
 	 	         all = 0;
                         break;
                    }
                    varptr = varptr-> next;
                  }
                  if (all == 1) {
	 	     return 1;
                  }

	       } /* check mib or scalar */

            } /* check protocol op */

         } /* match authenUser */

     } /* for each configured authenUser profile */

     /* didnt find match, sorry */
     return 0;
}

/* ************************************************************* */
